#
## HCA-TO-SCEA helper functions and maps.
#

import glob
import re
import sys

import requests as rq
from itertools import chain
import copy

from os.path import splitext, basename

import pandas as pd

def convert_to_snakecase(label):
    return re.sub(r'(\s-\s)|\s', '_', label).lower()

def reformat_value(sheet_dict, sheet, col_name):

    if col_name == "project.publications.pmid":
        values = list(sheet_dict[sheet][col_name].fillna('').replace(r'[\n\r]', ' ', regex=True))[0]
        if isinstance(values,float):
            values = str(int(values))
        else:
            values = values
    else:
        values = list(sheet_dict[sheet][col_name].fillna('').replace(r'[\n\r]', ' ', regex=True))

    return values

def get_tab_separated_list(sheet_dict, sheet, col_name, func=lambda x: x):
    tab = '\t'
    return tab.join([func(p) for p in reformat_value(sheet_dict, sheet, col_name)])

def get_first_letter(str):
    return str[0] if len(str) else ''

# Fetch all spreadsheet csv in a dir.
def get_all_spreadsheets(work_dir):
    file_names = glob.glob(f"{work_dir}/*.csv")
    file_names = [x for x in file_names if not 'big_table.csv' in x]

    spreadsheets = {}

    for file_name in file_names:
        spreadsheets[convert_to_snakecase(splitext(basename(file_name))[0])] = file_name

    for name, file_name in spreadsheets.items():
        newSheet = pd.read_csv(file_name, header=0, sep=";", skiprows=[0,1,2,4])
        newSheet = newSheet.applymap(str)
        newSheet = newSheet.applymap(lambda x: x.strip())
        spreadsheets[name] = newSheet.loc[:, ~newSheet.columns.str.contains('^Unnamed')]

    return spreadsheets

# Extract lists of protocols
# Helpers to convert lists in HCA spreadsheets (items are separated with two
# pipes `||`) to python lists.
def splitlist(list_):
    split_data = []

    try:
        if list_ != "nan":
            split_data = list_.split('||')
    except:
        pass

    return split_data

def save_files(work_dir, idf_file_name, sdrf_file_name, idf_file_contents, sdrf_file_contents):

    print(f"saving {work_dir}/{idf_file_name}")
    with open(f"{work_dir}/{idf_file_name}", "w") as idf_file:
        idf_file.write(idf_file_contents)

    '''Write the new sdrf file to a file.'''
    if not sdrf_file_contents.empty:
        print(f"saving {work_dir}/{sdrf_file_name}")
        sdrf_file_contents.to_csv(f"{work_dir}/{sdrf_file_name}", sep="\t", index=False)

def save_files_zip(zip_file,work_dir,idf_file_contents,sdrf_file_contents,idf_file_name,sdrf_file_name):

    print(f"saving {work_dir}/{idf_file_name}")
    with open(f"{work_dir}/{idf_file_name}", "w") as idf_file:
        idf_file.write(idf_file_contents)
    zip_file.write(f"{work_dir}/{idf_file_name}")

    if not sdrf_file_contents.empty:
        print(f"saving {work_dir}/{sdrf_file_name}")
        sdrf_file_contents.to_csv(f"{work_dir}/{sdrf_file_name}", sep="\t", index=False)
        zip_file.write(f"{work_dir}/{sdrf_file_name}")






import argparse
import json
import os
import sys
import pandas as pd
import copy
import zipfile

from hca_to_scea.helpers import multitab_excel_to_single_txt
from hca_to_scea.helpers import get_protocol_map
from hca_to_scea.helpers import fetch_fastq_path
from hca_to_scea.helpers import utils
from hca_to_scea.helpers import check_experimental_design
from hca_to_scea.helpers import split_dataset

pd.options.mode.chained_assignment = None

def rename_technology_type(technology_type, technology_dict):

    json_file = technology_dict[technology_type]

    return json_file

def get_person_roles(xlsx_dict):

    person_roles = utils.reformat_value(xlsx_dict, "project_contributors", "project.contributors.project_role.text")
    person_roles_submitter = utils.reformat_value(xlsx_dict, "project_contributors",
                                                  "project.contributors.corresponding_contributor")

    for (i, elem) in enumerate(person_roles_submitter):
        person_roles[i] = person_roles[i].lower()
        if elem == "yes":
            person_roles[i] = "submitter"
        elif elem == "no":
            if "curator" in person_roles[i]:
                person_roles[i] = "data curator"
            else:
                person_roles[i] = ""

    return person_roles

def get_author_list(xlsx_dict):

    authors = utils.reformat_value(xlsx_dict, "project_publications", "project.publications.authors")[0]
    author_list = authors.replace("||",", ")

    return author_list

def generate_idf_file(args, dataset_protocol_map, xlsx_dict, accession,
                      sdrf_file_name, related_scea_accessions):

    tab = '\t'
    person_roles = get_person_roles(xlsx_dict)
    #secondary_accessions = get_secondary_accessions(tracking_sheet, args) TBD from ingest.
    protocol_fields = get_protocol_map.get_idf_file_protocol_fields(dataset_protocol_map)
    author_list = get_author_list(xlsx_dict)

    related_scea_accessions = split_dataset.get_related_scea_accessions(accession, related_scea_accessions)

    if related_scea_accessions:

        idf_file_contents = f"""\

MAGE-TAB Version\t1.1
Investigation Title\t{utils.reformat_value(xlsx_dict, "project", "project.project_core.project_title")[0]}
Comment[Submitted Name]\t{utils.reformat_value(xlsx_dict, "project", "project.project_core.project_short_name")[0]}
Experiment Description\t{utils.reformat_value(xlsx_dict, "project", "project.project_core.project_description")[0]}
Public Release Date\t{args.public_release_date}
Person First Name\t{utils.get_tab_separated_list(xlsx_dict, "project_contributors", "project.contributors.name", lambda x: x.split(',')[0])}
Person Last Name\t{utils.get_tab_separated_list(xlsx_dict, "project_contributors", "project.contributors.name", lambda x: x.split(',')[2])}
Person Mid Initials\t{utils.get_tab_separated_list(xlsx_dict, "project_contributors", "project.contributors.name", lambda x: utils.get_first_letter(x.split(',')[1]))}
Person Email\t{utils.get_tab_separated_list(xlsx_dict, "project_contributors", "project.contributors.email")}
Person Affiliation\t{utils.get_tab_separated_list(xlsx_dict, "project_contributors", "project.contributors.institution")}
Person Address\t{utils.get_tab_separated_list(xlsx_dict, "project_contributors", "project.contributors.address")}
Person Roles\t{tab.join(person_roles)}
Protocol Type\t{tab.join([field[0] for field in protocol_fields])}
Protocol Name\t{tab.join([field[1] for field in protocol_fields])}
Protocol Description\t{tab.join([field[2] for field in protocol_fields])}
Protocol Hardware\t{tab.join([field[3] for field in protocol_fields])}
Term Source Name\tEFO\tArrayExpress
Term Source File\thttp://www.ebi.ac.uk/efo/efo.owl\thttp://www.ebi.ac.uk/arrayexpress/
Comment[AEExperimentType]\tRNA-seq of coding RNA from single cells
Experimental Factor Name\t{tab.join(args.experimental_factors)}
Experimental Factor Type\t{tab.join(args.experimental_factors)}
Comment[EAAdditionalAttributes]\t{''}
Comment[EACurator]\t{tab.join(args.curators)}
Comment[EAExpectedClusters]\t
Comment[ExpressionAtlasAccession]\t{accession}
Comment[RelatedExperiment]\t{tab.join(related_scea_accessions)}
Comment[HCALastUpdateDate]\t{args.hca_update_date}
Comment[SecondaryAccession]\t{args.project_uuid}
Comment[EAExperimentType]\t{args.experiment_type}
SDRF File\t{sdrf_file_name}

Investigation Title\t{utils.reformat_value(xlsx_dict, "project_publications", "project.publications.title")[0]}
Publication Author List\t{author_list}
PubMed ID\t{utils.reformat_value(xlsx_dict, "project_publications", "project.publications.doi")}
Publication DOI\t{utils.reformat_value(xlsx_dict, "project_publications", "project.publications.doi")[0]}
"""
    else:

        idf_file_contents = f"""\

MAGE-TAB Version\t1.1
Investigation Title\t{utils.reformat_value(xlsx_dict, "project", "project.project_core.project_title")[0]}
Comment[Submitted Name]\t{utils.reformat_value(xlsx_dict, "project", "project.project_core.project_short_name")[0]}
Experiment Description\t{utils.reformat_value(xlsx_dict, "project", "project.project_core.project_description")[0]}
Public Release Date\t{args.public_release_date}
Person First Name\t{utils.get_tab_separated_list(xlsx_dict, "project_contributors", "project.contributors.name", lambda x: x.split(',')[0])}
Person Last Name\t{utils.get_tab_separated_list(xlsx_dict, "project_contributors", "project.contributors.name", lambda x: x.split(',')[2])}
Person Mid Initials\t{utils.get_tab_separated_list(xlsx_dict, "project_contributors", "project.contributors.name", lambda x: utils.get_first_letter(x.split(',')[1]))}
Person Email\t{utils.get_tab_separated_list(xlsx_dict, "project_contributors", "project.contributors.email")}
Person Affiliation\t{utils.get_tab_separated_list(xlsx_dict, "project_contributors", "project.contributors.institution")}
Person Address\t{utils.get_tab_separated_list(xlsx_dict, "project_contributors", "project.contributors.address")}
Person Roles\t{tab.join(person_roles)}
Protocol Type\t{tab.join([field[0] for field in protocol_fields])}
Protocol Name\t{tab.join([field[1] for field in protocol_fields])}
Protocol Description\t{tab.join([field[2] for field in protocol_fields])}
Protocol Hardware\t{tab.join([field[3] for field in protocol_fields])}
Term Source Name\tEFO\tArrayExpress
Term Source File\thttp://www.ebi.ac.uk/efo/efo.owl\thttp://www.ebi.ac.uk/arrayexpress/
Comment[AEExperimentType]\tRNA-seq of coding RNA from single cells
Experimental Factor Name\t{tab.join(args.experimental_factors)}
Experimental Factor Type\t{tab.join(args.experimental_factors)}
Comment[EAAdditionalAttributes]
Comment[EACurator]\t{tab.join(args.curators)}
Comment[EAExpectedClusters]\t
Comment[ExpressionAtlasAccession]\t{accession}
Comment[HCALastUpdateDate]\t{args.hca_update_date}
Comment[SecondaryAccession]\t{args.project_uuid}
Comment[EAExperimentType]\t{args.experiment_type}
SDRF File\t{sdrf_file_name}

Investigation Title\t{utils.reformat_value(xlsx_dict, "project_publications", "project.publications.title")[0]}
Publication Author List\t{author_list}
PubMed ID\t{utils.reformat_value(xlsx_dict, "project_publications", "project.publications.pmid")}
Publication DOI\t{utils.reformat_value(xlsx_dict, "project_publications", "project.publications.doi")[0]}
"""

    return idf_file_contents

def reformat_age(age_list):

    updated_age_list = []

    for age in age_list:
        age = str(age)
        if ' - ' in age:
            age = age.replace('-', 'to')
        elif '-' in age and ' ' not in age:
            age = age.replace('-', ' to ')
        else:
            age = age

        updated_age_list.append(age)

    return updated_age_list


def get_sample_name_key(args, df):

    if args.name == 'cs_name':
        sample_name_key = 'cell_suspension.biomaterial_core.biomaterial_name'
    elif args.name == 'cs_id':
        sample_name_key = 'cell_suspension.biomaterial_core.biomaterial_id'
    elif args.name == 'sp_name':
        sample_name_key = 'specimen_from_organism.biomaterial_core.biomaterial_name'
    elif args.name == 'sp_id':
        sample_name_key = 'specimen_from_organism.biomaterial_core.biomaterial_id'

    return sample_name_key


def get_values_from_df(df, column):
    return df[column] if column in df.columns else ['']*df.shape[0]

def add_sequence_paths(sdrf, args):

    run_accessions = list(sdrf['Comment[ENA_RUN]'])

    try:
        sra_paths = fetch_fastq_path.get_sra_path_from_ena(args.study, run_accessions)
    except:
        sra_paths = fetch_fastq_path.get_sra_path_from_sra(args.study, run_accessions)
        if not sra_paths:
            try:
                sra_paths = fetch_fastq_path.get_sra_path_from_sra(run_accessions)
            except:
                sra_paths = fetch_fastq_path.get_sra_path_from_ena(args.study, run_accessions)

    if sra_paths:
        read1_names = []
        read2_names = []
        sra_names = []
        for key in sra_paths.keys():
            read1_names.append(key + "_1.fastq.gz")
            read2_names.append(key + "_2.fastq.gz")
            sra_names.append(sra_paths[key]['files'][0])
        sdrf['Comment[read1 file]'] = read1_names
        sdrf['Comment[read2 file]'] = read2_names
        sdrf['Comment[SRA_URI]'] = sra_names
    else:
        print("Could not find paths to SRA objects.")
        sdrf['Comment[SRA_URI]'] = 'PATH NOT FOUND'
        sdrf['Comment[read1 file]'] = 'PATH NOT FOUND'
        sdrf['Comment[read2 file]'] = 'PATH NOT FOUND'

    return sdrf


def get_new_protocol_column_names(sdrf, counter):

    new_column_names = []
    for col in sdrf.columns:
        column_name = "Protocol REF" + "_" + str(counter)
        counter += 1
        new_column_names.append(column_name)

    return new_column_names, counter

def order_protocols(protocols_sdrf_before_sequencing):
    collection_protocol_cols = []
    dissociation_protocol_cols = []
    enrichment_protocol_cols = []
    differentiation_protocol_cols = []
    library_preparation_protocol_cols = []
    for col in list(protocols_sdrf_before_sequencing.columns):
        if col.split(".protocol_core.protocol_id")[0] == "collection_protocol":
            collection_protocol_cols.append(col)
        if col.split(".protocol_core.protocol_id")[0] == "dissociation_protocol":
            dissociation_protocol_cols.append(col)
        if col.split(".protocol_core.protocol_id")[0] == "enrichment_protocol":
            enrichment_protocol_cols.append(col)
        if col.split(".protocol_core.protocol_id")[0] == "differentiation_protocol":
            differentiation_protocol_cols.append(col)
        if col.split(".protocol_core.protocol_id")[0] == "library_preparation_protocol":
            library_preparation_protocol_cols.append(col)
    new_columns = [sorted(collection_protocol_cols),sorted(dissociation_protocol_cols),sorted(enrichment_protocol_cols),sorted(differentiation_protocol_cols),sorted(library_preparation_protocol_cols)]
    new_columns = [item for sublist in new_columns for item in sublist]
    protocols_sdrf_before_sequencing = protocols_sdrf_before_sequencing[new_columns]
    return protocols_sdrf_before_sequencing

def add_protocol_columns(df, dataset_protocol_map):

    def convert_term(term, name):
        return get_protocol_map.map_proto_to_id(term, dataset_protocol_map)

    def convert_row(row):
        return row.apply(lambda x: convert_term(x, row.name))

    protocols_list_before_sequencing = ['collection_protocol', 'dissociation_protocol', 'enrichment_protocol', 'differentiation_protocol', 'library_preparation_protocol']

    protocols_sdrf_before_sequencing = df[[col for (proto_type, cols) in get_protocol_map.map_of_hca_protocol_type_id_keys.items() if proto_type in
                 protocols_list_before_sequencing for col in cols]]

    protocols_sdrf_before_sequencing.fillna("", inplace=True)
    pd.set_option('display.max_columns', None)

    protocols_sdrf_before_sequencing = protocols_sdrf_before_sequencing.apply(convert_row)
    pd.set_option('display.max_columns', None)

    protocols_sdrf_before_sequencing = order_protocols(protocols_sdrf_before_sequencing)

    counter = 1
    new_column_names, counter = get_new_protocol_column_names(protocols_sdrf_before_sequencing, counter)
    protocols_sdrf_before_sequencing.columns = new_column_names

    protocols_sdrf_before_sequencing.fillna(value='', inplace=True)

    protocols_list_from_sequencing = ['sequencing_protocol']
    protocols_sdrf_from_sequencing = df[[col for (proto_type, cols) in get_protocol_map.map_of_hca_protocol_type_id_keys.items() if proto_type in protocols_list_from_sequencing for col in cols]]
    protocols_sdrf_from_sequencing = protocols_sdrf_from_sequencing.apply(convert_row)
    new_column_names, counter = get_new_protocol_column_names(protocols_sdrf_from_sequencing, counter)
    protocols_sdrf_from_sequencing.columns = new_column_names

    return protocols_sdrf_before_sequencing, protocols_sdrf_from_sequencing


def add_scea_specimen_columns(args, df, experimental_design):

    if experimental_design == "standard":

        '''Open dictionary of mapped hca2scea key:pairs for specimen metadata.'''
        with open(f"json_files/sdrf_map.json") as sdrf_map_file:
            sdrf_map = json.load(sdrf_map_file)

    else:

        if experimental_design == "cell_line_only":
            '''Open dictionary of mapped hca2scea key:pairs for specimen metadata.'''
            with open(f"json_files/sdrf_map_cell_line.json") as sdrf_map_file:
                sdrf_map = json.load(sdrf_map_file)

        elif experimental_design == "organoid":
            '''Open dictionary of mapped hca2scea key:pairs for specimen metadata.'''
            with open(f"json_files/sdrf_map_cell_line_organoid.json") as sdrf_map_file:
                sdrf_map = json.load(sdrf_map_file)

        else:
            '''Open dictionary of mapped hca2scea key:pairs for specimen metadata.'''
            with open(f"json_files/sdrf_map_organoid.json") as sdrf_map_file:
                sdrf_map = json.load(sdrf_map_file)

    '''Get user-specified HCA sample names key.'''
    sample_name_key = get_sample_name_key(args, df)

    '''Update dictionary of mapped hca2scea key:pairs with sample name key.'''
    sdrf_map.update({
        'Source Name': sample_name_key,
        'Assay Name': sample_name_key,
        'Scan Name': sample_name_key,
        'Extract Name': sample_name_key,
        'Comment[BioSD_SAMPLE]': 'cell_suspension.biomaterial_core.biosamples_accession',
        'Comment[ENA_EXPERIMENT]': 'cell_suspension.insdc_experiment.insdc_experiment_accession',
        'Comment[ENA_RUN]': 'sequence_file.insdc_run_accessions',
        'Comment[technical replicate group]': 'cell_suspension.biomaterial_core.biosamples_accession'
        })

    '''Extract the HCA metadata values using the HCA keys in sdrf_map.'''
    sdrf = pd.DataFrame({k: get_values_from_df(df, v) for k, v in sdrf_map.items()})
    sdrf = sdrf.fillna('')

    '''Reformat the age values so they are excel-friendly.'''
    sdrf['Characteristics[age]'] = reformat_age(list(sdrf['Characteristics[age]']))

    '''Reformat living status to align with SCEA controlled vocabulary.'''
    sdrf['Characteristics[organism status]'] = sdrf['Characteristics[organism status]'].apply(lambda x: 'alive' if x.lower() in ['yes', 'y'] else 'dead')

    '''Add Material Type using SCEA controlled vocabulary.'''
    sdrf['Material Type_1'] = 'cell'

    return sdrf

def generate_sdrf_file(args, df, xlsx_dict, dataset_protocol_map, experimental_design, technology_dict):

    '''Generate a dataframe with SCEA specimen metadata.'''
    sdrf_1 = add_scea_specimen_columns(args, df, experimental_design)

    '''Get technology-specific SCEA metadata and add to sdrf_1 dataframe.'''
    technology_type = list(xlsx_dict["library_preparation_protocol"]["library_preparation_protocol.library_construction_method.ontology_label"])[0]
    technology_type = rename_technology_type(technology_type,technology_dict)
    try:
        with open(f"json_files/{technology_type}.json") as technology_json_file:
            technology_type_dict = json.load(technology_json_file)
    except:
        print("Technology type {} is not yet supported. Please ask Ami to add it to the technology type map.".format(technology_type))
        sys.exit()
    for key in technology_type_dict.keys():
        sdrf_1[key] = technology_type_dict[key]

    '''Edit single cell isolation method if the user-specified that FACS was used.'''
    if args.facs is True:
        sdrf_1['Comment[single_cell_isolation]'] = 'FACS'*sdrf_1.shape[0]
    # To do: add 'Comment[single_cell_isolation]' == Fluidigm C1 for Fluidigm C1 experiments.

    '''To Do. Get immunophenotype and treatment/stimulus information, if there is any.'''
    #"Characteristics[immunophenotype]":"enrichment_protocol.markers"
    sdrf_1['Characteristics[immunophenotype]'] = '' * sdrf_1.shape[0]
    #"Characteristics[stimulus]":"cell_suspension.growth_conditions.drug_treatment","differentiation_protocol.small_molecules",
    sdrf_1['Characteristics[stimulus]'] = '' * sdrf_1.shape[0]

    '''Get the SRA object file names and file paths from SRA or ENA (try both).'''
    sdrf_2 = add_sequence_paths(sdrf_1, args)

    '''Check all required column names are present and reorder columns by SCEA defined order.'''

    with open(f"json_files/expected_columns.json", "r") as expected_columns_file:
        expected_columns_dict = json.load(expected_columns_file)

    if experimental_design == 'standard':
        expected_columns_ordered = expected_columns_dict['standard']
    elif experimental_design == 'organoid_only':
        expected_columns_ordered = expected_columns_dict['organoid_only']
    else:
        expected_columns_ordered = expected_columns_dict['cell_line']

    column_check = [col for col in expected_columns_ordered if col not in sdrf_2.columns]
    if column_check:
        print("Error: one or more expected columns is missing from sdrf.")
        print(column_check)
        sys.exit()
    else:
        sdrf_3 = sdrf_2[expected_columns_ordered]

    '''Add protocol columns with protocol metadata in the pre-defined SCEA column order.'''
    protocols_sdrf_before_sequencing, protocols_sdrf_from_sequencing = add_protocol_columns(df, dataset_protocol_map)

    idx = 18
    for col in protocols_sdrf_before_sequencing.columns:
        sdrf_3.insert(idx, col, list(protocols_sdrf_before_sequencing[col]))
        idx += 1

    idx = list(sdrf_3.columns).index('Comment[LIBRARY_SELECTION]') + 1
    for col in protocols_sdrf_from_sequencing.columns:
        sdrf_3.insert(idx, col, list(protocols_sdrf_from_sequencing[col]))
        idx += 1

    '''Fix Material Type and Protocol REF column names to SCEA standard columns.'''
    for col_name in sdrf_3.columns:
        if 'Protocol REF' in str(col_name):
            sdrf_3 = sdrf_3.rename(columns={col_name: "Protocol REF"})
        elif 'Material Type' in str(col_name):
            sdrf_3 = sdrf_3.rename(columns={col_name: "Material Type"})

    '''Remove empty columns if columns are optional.'''
    with open(f"json_files/optional_columns.json", "r") as optional_columns_file:
        optional_columns_dict = json.load(optional_columns_file)

    if experimental_design == 'standard':
        optional_columns = optional_columns_dict['standard']
    else:
        optional_columns = optional_columns_dict['cell_line']

    for column_name in optional_columns:
        if set(list(sdrf_3[column_name])) == {''}:
            sdrf_3 = sdrf_3.drop([column_name], axis=1)

    return sdrf_3

def create_magetab(xlsx_dict, dataset_protocol_map, df, args, experimental_design, accession_number, related_scea_accessions, technology_dict):

    accession = f"E-HCAD-{accession_number}"

    idf_file_name = f"{accession}.idf.txt"
    sdrf_file_name = f"{accession}.sdrf.txt"

    idf_file_contents = generate_idf_file(args, dataset_protocol_map, xlsx_dict, accession,
                                          sdrf_file_name, related_scea_accessions)

    sdrf_file_contents = generate_sdrf_file(args, df, xlsx_dict, dataset_protocol_map, experimental_design,
                                            technology_dict)

    return idf_file_contents, sdrf_file_contents, idf_file_name, sdrf_file_name

def main():
    parser = argparse.ArgumentParser(description="run hca -> scea tool")
    parser.add_argument(
        "-s",
        "--spreadsheet",
        type=str,
        required=True,
        help="Please provide a path to the HCA project spreadsheet."
    )
    parser.add_argument(
        "-id",
        "--project_uuid",
        type=str,
        required=True,
        help="Please provide an HCA ingest project submission id."
    )
    parser.add_argument(
        "-study",
        type=str,
        required=True,
        help="Please provide the SRA or ENA study accession."
    )
    parser.add_argument(
        "-name",
        type=str,
        required=False,
        default = 'cs_id',
        choices = ['cs_name','cs_id','sp_name','sp_id','other'],
        help="Please indicate which field to use as the sample name. cs=cell suspension, sp = specimen."
    )
    parser.add_argument(
        "-ac",
        "--accession_number",
        type=int,
        required=True,
        help="Provide an E-HCAD accession number. Please find the next suitable accession number by checking the google tracker sheet."
    )
    parser.add_argument(
        "-c",
        "--curators",
        nargs='+',
        required=True,
        help="space separated names of curators"
    )
    parser.add_argument(
        "-et",
        "--experiment_type",
        type=str,
        required=True,
        choices=['baseline','differential'],
        help="Please indicate whether this is a baseline or differential experimental design"
    )
    parser.add_argument(
        "-facs",
        action="store_true",
        required=False,
        help="Please specify if FACS was used to isolate single cells."
    )
    parser.add_argument(
        "-f",
        "--experimental_factors",
        nargs='+',
        required=True,
        help="space separated list of experimental factors"
    )
    parser.add_argument(
        "-pd",
        "--public_release_date",
        type=str,
        required=True,
        help="Please enter the public release date in this format: YYYY-MM-DD"
    )
    parser.add_argument(
        "-hd",
        "--hca_update_date",
        type=str,
        required=True,
        help="Please enter the last time the HCA prohect submission was updated in this format: YYYY-MM-DD"
    )
    parser.add_argument(
        "-o",
        "--output_dir",
        required=False,
        help="Provide full path to preferred output dir"
    )
    parser.add_argument(
        "-zip",
        "--zip_format",
        action="store_true",
        required=False,
        help="Please indicate whether you would like the script to output all"
             "txt files separately or together in 1 zip file."
    )

    args = parser.parse_args()
    if not args.output_dir:
        work_dir = f"script_spreadsheets/{os.path.splitext(os.path.basename(args.spreadsheet))[0]}"
    else:
        work_dir = args.output_dir

    if args.zip_format:
        zip_file = zipfile.ZipFile('hca-to-scea.zip', 'w')
    else:
        zip_file = None

    '''Merge the multitab spreadsheet into a single dataframe, while preserving the relationships
    between HCA biomaterials and protocols.
    '''
    xlsx_dict = multitab_excel_to_single_txt.multitab_excel_to_dict(work_dir, args.spreadsheet)

    '''Remove unused protocol tabs and corresponding protocol id column.'''
    xlsx_dict_tmp2 = multitab_excel_to_single_txt.remove_unused_protocols(xlsx_dict)

    '''Check whether multiple library preparation protocol technology types or 10X versions were
    used. If so, split xlsx_dict into a list of dicts separated by the technology type. Then,
    create idf and sdrf files for each of the dicts.'''

    technology_dict = {
        "Fluidigm C1-based library preparation": "smart-like",
        "10X 3' v1": "10Xv1_3",
        "10X 5' v1": "10Xv1_5",
        "10X 3' v2": "10Xv2_3",
        "10X 5' v2": "10Xv2_5",
        "10X 3' v3": "10Xv3_3",
        "10X 3' v1 sequencing": "10Xv1_3",
        "10X 5' v1 sequencing": "10Xv1_5",
        "10X 3' v2 sequencing": "10Xv2_3",
        "10X 5' v2 sequencing": "10Xv2_5",
        "10x 3' v3 sequencing": "10Xv3_3",
        "Drop-seq": "drop-seq",
        "inDrop": "drop-seq",
        "Smart-like": "smart-like",
        "Smart-seq2": "smart-seq",
        "Smart-seq": "smart-seq"
    }

    list_xlsx_dict = split_dataset.split_metadata_by_technology(xlsx_dict_tmp2,technology_dict)

    if len(list_xlsx_dict) > 1:
        accession_number_idx = [i for i in range(0,len(list_xlsx_dict))]
        accession_number_list = [int(args.accession_number) + i for i in accession_number_idx]

    else:
        accession_number_idx = [0]
        accession_number_list = [args.accession_number]

    for i in range(0,len(list_xlsx_dict)):

        xlsx_dict = list_xlsx_dict[i]
        accession_number = accession_number_list[i]

        related_scea_accessions_idx = copy.deepcopy(accession_number_idx)
        related_scea_accessions_idx.remove(i)
        related_scea_accessions = [accession_number_list[i] for i in related_scea_accessions_idx]

        '''Run checks to see whether the experimental design is compatibile and save the experimental
        design type in a variable for later.'''

        ''' Get the experimental design '''
        experimental_design = check_experimental_design.get_experimental_design(xlsx_dict)

        ''' Check if samples are pooled '''
        pooled_samples = check_experimental_design.check_for_pooled_samples(xlsx_dict)
        if pooled_samples:
            print("The hca-to-scea tool does not support pooled donors or pooled samples."
                "The dataset should be curated manually.")
            sys.exit()

        '''The merged df consists of a row per read index (read1, read2, index1). To conform to
        SCEA MAGE-TAB format, the rows should be merged so that there is 1 row per unique run accession.
        '''
        merged_df = multitab_excel_to_single_txt.merge_dataframes(xlsx_dict,experimental_design)
        merged_df_unique_runs = merged_df.drop_duplicates(subset=['sequence_file.insdc_run_accessions'])
        clean_merged_df = multitab_excel_to_single_txt.clean_df(merged_df_unique_runs)

        '''Extract the list of unique protocol ids from protocol types which can have more than one instance and
        creates extra columns in the df for each of the ids.'''
        df = multitab_excel_to_single_txt.create_new_protocol_columns(clean_merged_df, xlsx_dict, experimental_design)

        '''Create a map between the HCA protocol id and a new assigned SCEA protocol id. Use it to store the
        key protocol metadata that will be added to the SCEA sdrf file.'''
        dataset_protocol_map = get_protocol_map.prepare_protocol_map(xlsx_dict, df, args)

        '''Refactoring of the below TBD.'''
        idf_file_contents, sdrf_file_contents, idf_file_name, sdrf_file_name = create_magetab(xlsx_dict, dataset_protocol_map, df, args, experimental_design, accession_number,related_scea_accessions,technology_dict)

        if not args.zip_format:
            utils.save_files(work_dir, idf_file_name, sdrf_file_name, idf_file_contents, sdrf_file_contents)
        else:
            utils.save_files_zip(zip_file, work_dir, idf_file_contents, sdrf_file_contents, idf_file_name,
                                 sdrf_file_name)

    if zip_file:
        zip_file.close()

if __name__ == '__main__':
    main()



import logging
import os
import sys
import unittest
import pandas as pd
import numpy as np
from pandas.testing import assert_frame_equal


class CharacteristicTest(unittest.TestCase):

    def setUp(self):
        self.verificationErrors = {}
        self.output_base = 'output/'
        if not sys.warnoptions:
            import warnings
            warnings.simplefilter("ignore")

    def test_hca2scea_characteristic(self):
        # run tool
        arguments_df = pd.read_csv("test/golden/arguments.csv", comment='#')
        for i in range(0,arguments_df.shape[0]):
            spreadsheet = "test/golden/" + list(arguments_df['spreadsheet'])[i]
            with self.subTest(spreadsheet="test/golden/" + list(arguments_df['spreadsheet'])[i]):
                arguments = arguments_df.loc[arguments_df['spreadsheet'] == os.path.basename(spreadsheet)]
                output_dir = self.run_tool(spreadsheet, arguments)
                self.check_output(output_dir, spreadsheet)

    def get_file_content(self, file):
        if file.split(".")[-2] == 'sdrf':
            file_contents = self.load_sdrf_file(file)
        elif file.split(".")[-2] == 'idf':
            file_contents = self.load_idf_file(file)
            logging.info(file_contents.head())
        elif file.split(".")[-1] == 'csv':
            file_contents = self.load_big_table_file(file)
        else:
            raise ValueError(f'unsupported test file format: {file}')
        return file_contents

    def load_big_table_file(self, file):
        return pd.read_csv(file, sep=';')

    def load_idf_file(self, file):
        return pd.read_csv(file,
                           sep='^([^\t]+)\t',
                           engine='python',
                           usecols=[0, 1, 2],
                           names=['idx', 'name', 'value'])

    def load_sdrf_file(self, file):
        return pd.read_csv(file, sep='\t')

    def assert_dataframes_equal(self, golden_contents, output_contents, tag=None):
        difference_locations = golden_contents != output_contents
        changed_from = golden_contents[difference_locations].dropna(how='all')
        changed_to = output_contents[difference_locations].dropna(how='all')
        diff:pd.DataFrame = changed_from.join(changed_to,
                                              lsuffix='_expected', rsuffix='_actual',
                                              sort=False)

        diff = diff.melt()
        if len(diff) != 0:
            diff_file = f'{self.output_base}/diff.html'
            if tag:
                diff_file = f'{self.output_base}/diff-{tag}.html'
            diff.to_html(diff_file)
        assert len(diff) == 0, f'diffs found comparing {tag}\n{diff.to_string()}'

    def check_equal_lines(self, golden_contents, output_contents, msg=None):
        self.assertMultiLineEqual(golden_contents,output_contents, msg)

    def check_output(self, output_dir, spreadsheet):
        golden_output_dir = 'test/golden/expected/' + os.path.basename(spreadsheet).split(".xlsx")[0]
        for golden_file in os.listdir(golden_output_dir):
            golden_file_basename = os.path.basename(golden_file)
            output_file = os.path.join(output_dir, golden_file_basename)
            golden_contents = self.get_file_content(os.path.join(golden_output_dir, golden_file_basename))
            output_contents = self.get_file_content(output_file)
            try:
                if isinstance(golden_contents, pd.DataFrame):
                    self.assert_dataframes_equal(golden_contents, output_contents, tag=golden_file_basename)
                else:
                    self.check_equal_lines(golden_contents, output_contents, f'diffs found comparing {golden_file_basename}')
            except Exception as e:
                raise AssertionError(f'problem with {golden_file}') from e
    def run_tool(self, spreadsheet, arguments):
        output_name = os.path.basename(spreadsheet).split(".xlsx")[0]
        output_dir = self.output_base + output_name
        arguments = arguments.reset_index()
        os.system(
            f'python3 hca2scea.py -s {spreadsheet} -o {output_dir} -id {arguments["HCA project uuid"][0]} -ac {arguments["E-HCAD accession"][0]} -c {arguments["curator initials"][0]} -et {arguments["experiment type"][0]} -f {arguments["factor values"][0]} -pd {arguments["public release date"][0]} -hd {arguments["hca last update date"][0]} -study {arguments["study accession"][0]}')
        return output_dir

if __name__ == '__main__':
    unittest.main()



import pandas as pd

'''Map of HCA protocol type names to SCEA protocol type names.'''
protocol_type_map = {
    'collection_protocol': "sample collection protocol",
    'dissociation_protocol': "enrichment protocol",
    '??????????????????????': "nucleic acid extraction protocol",
    'enrichment_protocol': "enrichment protocol",
    "differentiation_protocol": "growth protocol",
    'library_preparation_protocol': "nucleic acid library construction protocol",
    'sequencing_protocol': "nucleic acid sequencing protocol",
}

'''Order of SCEA protocol type names by experimental stage (SCEA requirement).'''
ordered_hca_protocol_types = [
    'collection_protocol',
    'dissociation_protocol',
    'enrichment_protocol',
    'differentiation_protocol',
    'library_preparation_protocol',
    'sequencing_protocol',
]

'''HCA columns names for HCA protocol type ids.'''
map_of_hca_protocol_type_id_keys = {
    'collection_protocol': ["collection_protocol.protocol_core.protocol_id"],
    'library_preparation_protocol': ["library_preparation_protocol.protocol_core.protocol_id"],
    'sequencing_protocol': ["sequencing_protocol.protocol_core.protocol_id"]
}

'''There can be multiple dissociation protocols and/or enrichment protocols
performed on a single specimen. This dict stores such protocol types.'''
multiprotocols = {
    'dissociation_protocol': "dissociation_protocol.protocol_core.protocol_id",
    'enrichment_protocol': "enrichment_protocol.protocol_core.protocol_id",
    'differentiation_protocol': "differentiation_protocol.protocol_core.protocol_id"
}

'''There can be multiple dissociation protocols and/or enrichment protocols
performed on a single specimen. The following 2 functions convert a lists of protocols
from an HCA spreadsheet, separated by '||', to a standard python list.'''
def splitlist(list_):

    split_data = []
    try:
        if list_ != "nan" and list_ != "":
            split_data = list_.split('||')
            split_data = sorted(split_data)
    except:
        pass

    return split_data

def split_multiprotocols(df, proto_column):

    proto_series = df[proto_column].apply(splitlist)
    proto_df = pd.DataFrame(proto_series.values.tolist())
    proto_df_columns = [f'{proto_column}_{y}' for y in range(len(proto_df.columns))]
    proto_df.columns = proto_df_columns
    proto_df[f'{proto_column}_count'] = proto_series.str.len()
    proto_df[f'{proto_column}_list'] = proto_series

    return (proto_df, proto_df_columns)

'''Given a list of unique protocol type ids, get the corresponding protocol type id
in SCEA's protocol accession format.'''
def get_scea_protocol_id(dataset_protocol_map, hca_protocol_type, hca_protocol_type_ids,
                         scea_protocol_accession, protocol_id_counter):

    for hca_protocol_type_id in hca_protocol_type_ids:
        if hca_protocol_type_id is not None:
            protocol_id_counter += 1
            scea_protocol_type_id = f"P-{scea_protocol_accession}-{protocol_id_counter}"
            dataset_protocol_map[hca_protocol_type].update({hca_protocol_type_id:
                                                                {'scea_id': scea_protocol_type_id,
                                                                 'hca_ids': [hca_protocol_type_id]}})

    return dataset_protocol_map, protocol_id_counter


'''For HCA protocol type in the ordered list of HCA protocols, get the list of unique
protocol type ids given the hca protocol type id key (column) for the HCA protocol type.'''
def get_unique_protocol_ids(df, map_of_hca_protocol_type_id_keys, hca_protocol_type):

    for (hca_ptype, hca_protocol_type_id_keys) in map_of_hca_protocol_type_id_keys.items():
        if hca_ptype == hca_protocol_type:
            hca_protocol_type_ids = []
            for hca_protocol_type_id_key in hca_protocol_type_id_keys:
                hca_protocol_type_ids = hca_protocol_type_ids + pd.unique(df[hca_protocol_type_id_key]).tolist()

    return hca_protocol_type_ids

'''Create a dataset protocol map (dictionary), containing types of HCA protocols and inside each,
 a map from HCA ids to SCEA ids.'''
def get_dataset_protocol_map(df, ordered_hca_protocol_types, map_of_hca_protocol_type_id_keys, scea_protocol_accession_prefix):

    '''First, we prepare an ID minter for the protocols following SCEA MAGE-TAB standards.'''
    protocol_id_counter = 0

    '''Create empty dictionary to store the HCA protocol ids mapped to the SCEA protocol ids.'''
    dataset_protocol_map = {x: {} for x in ordered_hca_protocol_types}

    for hca_protocol_type in ordered_hca_protocol_types:

        if hca_protocol_type in map_of_hca_protocol_type_id_keys.keys():

            '''For HCA protocol type in the ordered list of HCA protocols, get the list of unique
            protocol type ids given the hca protocol type id key (column) for the HCA protocol type.'''
            hca_protocol_type_ids = get_unique_protocol_ids(df, map_of_hca_protocol_type_id_keys, hca_protocol_type)

            '''Given a list of unique protocol type ids, get the corresponding protocol type id
            in SCEA's protocol accession format and update the dataset_protocol_map.'''
            dataset_protocol_map, protocol_id_counter = get_scea_protocol_id(dataset_protocol_map, hca_protocol_type, hca_protocol_type_ids, scea_protocol_accession_prefix, protocol_id_counter)

    return dataset_protocol_map

'''Extracts hca protocol metadata from the hca protocol dataframes in xlsx_dict.'''
def extract_protocol_metadata(
    dataset_protocol_map,
    xlsx_dict,
    column_to_extract,
    new_key,
    hca_protocol_types
):
    for hca_protocol_type, hca_protocol_type_dict in dataset_protocol_map.items():
        if hca_protocol_type in hca_protocol_types:
            for hca_protocol_type_id, hca_protocol_type_id_dict in hca_protocol_type_dict.items():
                extracted_data = xlsx_dict[hca_protocol_type].loc[xlsx_dict[hca_protocol_type][f'{hca_protocol_type}.protocol_core.protocol_id'] == hca_protocol_type_id][f'{hca_protocol_type}.{column_to_extract}'].tolist()
                if len(extracted_data):
                    hca_protocol_type_id_dict[new_key] = extracted_data[0]
                else:
                    hca_protocol_type_id_dict[new_key] = ''

# Get protocol types from protocol map.
def get_idf_file_protocol_fields(protocol_map):

    proto_types = [protocol_type_map[protocol_type] for (protocol_type, value) in protocol_map.items() for repeats in range(len(value.keys()))]
    proto_names = [protocol['scea_id'] for (protocol_type, protocols) in protocol_map.items() for (protocol_name, protocol) in protocols.items()]
    proto_descs = [protocol['description'] for (protocol_type, protocols) in protocol_map.items() for (protocol_name, protocol) in protocols.items()]
    proto_hware = [protocol.get('hardware', '') for (protocol_type, protocols) in protocol_map.items() for (protocol_name, protocol) in protocols.items()]

    return list(zip(proto_types, proto_names, proto_descs, proto_hware))

'''Maps a HCA protocol name to a SCEA ID.'''
def map_proto_to_id(protocol_name, protocol_map):
    for key,proto_type in protocol_map.items():
        for proto in proto_type.values():
            if protocol_name in proto['hca_ids']:
                return proto.get('scea_id')
    return ''

def prepare_protocol_map(xlsx_dict, df, args):

    scea_protocol_accession_prefix = f"HCAD{args.accession_number}"

    dataset_protocol_map = get_dataset_protocol_map(df,
                                                    ordered_hca_protocol_types,
                                                    map_of_hca_protocol_type_id_keys,
                                                    scea_protocol_accession_prefix)

    ''' Get the description text for all protocol types.'''
    extract_protocol_metadata(dataset_protocol_map, xlsx_dict,
                              f"protocol_core.protocol_description", "description",
                              ordered_hca_protocol_types)

    ''' Get the instrument type and hardware type for the sequencing protocol(s).'''
    extract_protocol_metadata(dataset_protocol_map, xlsx_dict,
                              f"instrument_manufacturer_model.ontology_label", "hardware",
                              ["sequencing_protocol"])

    return dataset_protocol_map


def get_experimental_design(xlsx_dict: {}):

    if 'specimen_from_organism' in xlsx_dict.keys():
        if xlsx_dict['specimen_from_organism'].empty:
            specimen = False
        else:
            specimen = True
    else:
        specimen = False

    if 'cell_line' in xlsx_dict.keys():
        if xlsx_dict['cell_line'].empty:
            cell_line = False
        else:
            cell_line = True
    else:
        cell_line = False

    if 'organoid' in xlsx_dict.keys():
        if xlsx_dict['organoid'].empty:
            organoid = False
        else:
            organoid = True
    else:
        organoid = False

    if specimen:

        if cell_line and not organoid:
            experimental_design = "cell_line_only"
        elif not cell_line and organoid:
            experimental_design = "organoid_only"
        elif cell_line and organoid:
            experimental_design = "organoid"
        else:
            experimental_design = "standard"

    return experimental_design

def check_for_pooled_samples(xlsx_dict):

    specimen_metadata = xlsx_dict["specimen_from_organism"]
    specimen_ids = list(specimen_metadata['specimen_from_organism.biomaterial_core.biomaterial_id'])
    pooled_samples_specimen = [specimen_id for specimen_id in specimen_ids if "||" in specimen_id]

    pooled_samples_cell_line = []
    if "cell_line" in xlsx_dict.keys() and not xlsx_dict["cell_line"].empty:
        cell_line_metadata = xlsx_dict["cell_line"]
        if 'cell_line.biomaterial_core.biomaterial_id' in cell_line_metadata.columns:
            cell_line_ids = list(cell_line_metadata['cell_line.biomaterial_core.biomaterial_id'])
            pooled_samples_cell_line = [cell_line_id for cell_line_id in cell_line_ids if "||" in cell_line_id]

    pooled_samples_organoid = []
    if "organoid" in xlsx_dict.keys() and not xlsx_dict["organoid"].empty:
        organoid_metadata = xlsx_dict["organoid"]
        if 'organoid.biomaterial_core.biomaterial_id' in organoid_metadata.columns:
            organoid_ids = list(organoid_metadata['organoid.biomaterial_core.biomaterial_id'])
            pooled_samples_organoid = [organoid_id for organoid_id in organoid_ids if "||" in organoid_id]

    donor_metadata = xlsx_dict["donor_organism"]
    donor_ids = list(donor_metadata['donor_organism.biomaterial_core.biomaterial_id'])
    pooled_samples_donor = [str(donor_id) for donor_id in donor_ids if "||" in str(donor_id)]

    if pooled_samples_specimen or pooled_samples_donor or pooled_samples_cell_line or pooled_samples_organoid:
        pooled_samples = True
    else:
        pooled_samples = False

    return pooled_samples

def check_technology_eligibility(library_method,technology_dict):

    eligible = library_method in technology_dict.keys()

    return eligible



def get_experiment_pkg_objects(xml_content):
    """Get experiment package objects from xml."""
    for experiment_package in xml_content.findall('EXPERIMENT_PACKAGE'):
        yield experiment_package



import copy
import pandas as pd

from hca_to_scea.helpers import check_experimental_design

def filter_biomaterials(xlsx_dict, xlsx_dict_tmp, library_protocol):

    xlsx_dict_tmp["sequence_file"] = xlsx_dict["sequence_file"].loc[
        xlsx_dict["sequence_file"]["library_preparation_protocol.protocol_core.protocol_id"] == library_protocol]
    cell_suspension_ids = xlsx_dict_tmp["sequence_file"]["cell_suspension.biomaterial_core.biomaterial_id"].values
    xlsx_dict_tmp["cell_suspension"] = xlsx_dict["cell_suspension"][
        xlsx_dict["cell_suspension"]["cell_suspension.biomaterial_core.biomaterial_id"].isin((cell_suspension_ids))]

    biomaterial_tabs = ["specimen_from_organism", "organoid", "cell_line"]
    for biomaterial_tab in biomaterial_tabs:
        if biomaterial_tab in xlsx_dict.keys():
            if not xlsx_dict_tmp[biomaterial_tab].empty:
                id_column = "{}.biomaterial_core.biomaterial_id".format(biomaterial_tab)
                if id_column in xlsx_dict_tmp["cell_suspension"].columns:
                    if not xlsx_dict_tmp["cell_suspension"][id_column].empty and not xlsx_dict_tmp["cell_suspension"][id_column].isna().all():
                        specimen_ids = xlsx_dict_tmp["cell_suspension"][
                            "{}.biomaterial_core.biomaterial_id".format(biomaterial_tab)].values
                        xlsx_dict_tmp[biomaterial_tab] = xlsx_dict[biomaterial_tab][
                            xlsx_dict[biomaterial_tab]["{}.biomaterial_core.biomaterial_id".format(biomaterial_tab)].isin(
                                (specimen_ids))]
                if "donor_organism.biomaterial_core.biomaterial_id" in xlsx_dict_tmp[biomaterial_tab].columns:
                    donor_ids_specimen = xlsx_dict_tmp[biomaterial_tab][
                        "donor_organism.biomaterial_core.biomaterial_id"].values
                    xlsx_dict_tmp["donor_organism"] = xlsx_dict["donor_organism"][
                        xlsx_dict["donor_organism"]["donor_organism.biomaterial_core.biomaterial_id"].isin(
                            (donor_ids_specimen))]

    return xlsx_dict_tmp

def filter_protocols(xlsx_dict_tmp):

    biomaterial_tabs = ["specimen_from_organism", "organoid", "cell_line", "cell_suspension", "sequence_file"]
    protocol_tabs = ["collection_protocol","dissociation_protocol","enrichment_protocol","differentiation_protocol","library_preparation_protocol","sequencing_protocol"]

    for protocol_tab in protocol_tabs:
        if protocol_tab in xlsx_dict_tmp.keys():
            id_column = "{}.protocol_core.protocol_id".format(protocol_tab)
            protocol_id_list = []
            for biomaterial_tab in biomaterial_tabs:
                if biomaterial_tab in xlsx_dict_tmp.keys():
                    if id_column in xlsx_dict_tmp[biomaterial_tab].columns:
                        protocol_id_list.extend(list(xlsx_dict_tmp[biomaterial_tab][id_column]))
            protocol_id_list = list(set(protocol_id_list))
            protocol_id_list = [protocol_id for protocol_id in protocol_id_list if pd.notnull(protocol_id)]
            protocol_id_split = [protocol_id.split("||") for protocol_id in protocol_id_list if "||" in protocol_id]
            if protocol_id_split:
                protocol_id_split = [item for sublist in protocol_id_split for item in sublist]
                protocol_id_split.extend([protocol_id for protocol_id in protocol_id_list if "||" not in protocol_id])
                protocol_id_list = protocol_id_split
            keep_protocols = [protocol_id for protocol_id in list(xlsx_dict_tmp[protocol_tab][id_column]) if protocol_id in protocol_id_list]
            xlsx_dict_tmp[protocol_tab] = xlsx_dict_tmp[protocol_tab][xlsx_dict_tmp[protocol_tab][id_column].isin((keep_protocols))]

    return xlsx_dict_tmp

def split_metadata_by_technology(xlsx_dict, technology_dict):

    '''Check for ineligible technoloy types and remove them from the library preparation protocol tab.'''
    for i in range(0,xlsx_dict["library_preparation_protocol"].shape[0]):
        library_protocol_technology = xlsx_dict["library_preparation_protocol"]["library_preparation_protocol.library_construction_method.ontology_label"][i]
        eligible = check_experimental_design.check_technology_eligibility(library_protocol_technology,technology_dict)
        if eligible == "False":
            xlsx_dict["library_preparation_protocol"].drop(xlsx_dict["library_preparation_protocol"].index[i])

    '''Get the list of eligible unique library preparation protocol types.'''
    if xlsx_dict["library_preparation_protocol"].empty:
        print("The library preparation technology type is not compatible. Please speak to Ami.")
        sys.exit()
    else:
        library_protocol_ids = xlsx_dict["library_preparation_protocol"]["library_preparation_protocol.protocol_core.protocol_id"].values
        list_xlsx_dict = []
        for library_protocol in library_protocol_ids:
            xlsx_dict_tmp = copy.deepcopy(xlsx_dict)
            xlsx_dict_tmp = filter_biomaterials(xlsx_dict,xlsx_dict_tmp,library_protocol)
            xlsx_dict_tmp2 = copy.deepcopy(xlsx_dict_tmp)
            xlsx_dict_tmp3 = filter_protocols(xlsx_dict_tmp2)
            list_xlsx_dict.append(xlsx_dict_tmp3)

    return list_xlsx_dict

def get_related_scea_accessions(accession, related_scea_accessions):

    if related_scea_accessions:
        related_scea_accessions = [str(accession.split("E-HCAD-")[0]) + str(related_scea_acc) for related_scea_acc in related_scea_accessions]
        if related_scea_accession:
            related_scea_accessions.extend(related_scea_accession)

    return related_scea_accessions


import argparse
import os
import requests as rq
from xml.etree import ElementTree
import multiprocessing
from contextlib import contextmanager
import pandas as pd

@contextmanager
def poolcontext(*args, **kwargs):
    pool = multiprocessing.Pool(*args, **kwargs)
    yield pool
    pool.terminate()

def parse_xml(xml_content):
    for experiment_package in xml_content.findall('EXPERIMENT_PACKAGE'):
        yield experiment_package

def filter_paths(sdrf, paths):
    runs = list(sdrf['Comment[ENA_RUN]'])
    read1_files = []
    read2_files = []
    index1_files = []
    index2_files = []
    read1_paths = []
    read2_paths = []
    index1_paths = []
    index2_paths = []
    sra_paths = []
    sra_read1_files = []
    sra_read2_files = []
    for i in range(0, len(runs)):
        run = runs[i]
        if paths[run]['filetype'] == 'fastq file':
            read1_files.append(paths[run]['filename_read1'])
            read1_paths.append(paths[run]['filepath_read1'])
            read2_files.append(paths[run]['filename_read2'])
            read2_paths.append(paths[run]['filepath_read2'])
            if 'filename_index1' in paths[run].keys():
                index1_files.append(paths[run]['filename_index1'])
                index1_paths.append(paths[run]['filepath_index1'])
            else:
                index1_files.append('')
                index1_paths.append('')
            if 'filename_index2' in paths[run].keys():
                index2_files.append(paths[run]['filename_index2'])
                index2_paths.append(paths[run]['filepath_index2'])
            else:
                index2_files.append('')
                index2_paths.append('')
        elif paths[run]['filetype'] == 'SRA file':
            sra_paths.append(paths[run]['filepath'])
            file1 = run + "_1.fastq.gz"
            file2 = run + "_2.fastq.gz"
            sra_read1_files.append(file1)
            sra_read2_files.append(file2)
        else:
            continue
    if read1_files:
        sdrf['Comment[read1 file]'] = read1_files
        sdrf['Comment[read1 FASTQ_URI]'] = read1_paths
        sdrf['Comment[read2 file]'] = read2_files
        sdrf['Comment[read2 FASTQ_URI]'] = read2_paths
        if index1_files:
            sdrf['Comment[index1 file]'] = index1_files
            sdrf['Comment[index1 FASTQ_URI]'] = index1_paths
        else:
            sdrf['Comment[index1 file]'] = '' * len(read1_files)
            sdrf['Comment[index1 FASTQ_URI]'] = '' * len(read1_files)
        sdrf['Comment[SRA_URI]'] = '' * len(read1_files)
    elif sra_paths:
        sdrf['Comment[SRA_URI]'] = sra_paths
        sdrf['Comment[read1 file]'] = sra_read1_files
        sdrf['Comment[read2 file]'] = sra_read2_files
        sdrf['Comment[index1 file]'] = ''*len(sra_read1_files)
        sdrf['Comment[read1 FASTQ_URI]'] = ''*len(sra_read1_files)
        sdrf['Comment[read2 FASTQ_URI]'] = ''*len(sra_read1_files)
        sdrf['Comment[index1 FASTQ_URI]'] = ''*len(sra_read1_files)

    sdrf.drop_duplicates(keep=False, inplace=True)
    return sdrf

def sort_sra(paths):
    paths_new = paths
    for accession in paths.keys():
        sra_path = paths[accession]['files'][0]
        paths_new[accession]['filepath'] = sra_path
        paths_new[accession]['filetype'] = 'SRA file'
    return paths_new

def sort_fastq(paths):
    invalid_fastq = 'False'
    paths_new = paths
    for accession in paths.keys():
        fastq = paths[accession]['files']
        for file in fastq:
            if '_R1' in file or '_R2' in file or '_R3' in file or '_R4' in file or '_I1' in file or '_I2' in file:
                if '_R1' in file:
                    paths_new[accession]['filename_read1'] = os.path.basename(file)
                    paths_new[accession]['filepath_read1'] = file
                if '_R2' in file:
                    paths_new[accession]['filename_read2'] = os.path.basename(file)
                    paths_new[accession]['filepath_read2'] = file
                if '_I1' in file or '_R3' in file:
                    paths_new[accession]['filename_index1'] = os.path.basename(file)
                    paths_new[accession]['filepath_index1'] = file
                if '_I2' in file or '_R4' in file:
                    paths_new[accession]['filename_index2'] = os.path.basename(file)
                    paths_new[accession]['filepath_index2'] = file
            elif '_1' in file or '_2' in file or '_3' in file or '_4' in file and '_R' not in file:
                if '_1' in file:
                    paths_new[accession]['filename_read1'] = os.path.basename(file)
                    paths_new[accession]['filepath_read1'] = file
                if '_2' in file:
                    paths_new[accession]['filename_read2'] = os.path.basename(file)
                    paths_new[accession]['filepath_read2'] = file
                if '_3' in file:
                    paths_new[accession]['filename_index1'] = os.path.basename(file)
                    paths_new[accession]['filepath_index1'] = file
                if '_4' in file:
                    paths_new[accession]['filename_index2'] = os.path.basename(file)
                    paths_new[accession]['filepath_index2'] = file
        paths_new[accession]['filetype'] = 'fastq file'
        if 'filename_read1' not in paths_new[accession].keys() or 'filename_read2' not in paths_new[accession].keys():
            invalid_fastq = 'True'
    if invalid_fastq == 'True':
        return {}
    else:
        return paths_new

def pool_retrieve_xml_from_sra(run_lists):
    try:
        result_list = []
        with poolcontext(processes=1) as pool:
            result_list.append(pool.map(retrieve_paths_from_sra, run_lists))
    except KeyboardInterrupt:
        print("Process has been interrupted.")
        pool.terminate()
    return result_list

def run_accessions_to_parts(run_accessions):
    for i in range(0, len(run_accessions), 100):
        yield run_accessions[i:i + 100]

def retrieve_xml_from_sra(run_accessions):
    url = f'https://eutils.ncbi.nlm.nih.gov/entrez/eutils/efetch/fcgi?db=sra&id={",".join(run_accessions)}'
    srr_metadata_url = rq.get(url)
    paths_sra = {}
    for accession in run_accessions:
        paths_sra[accession] = {}
        paths_sra[accession]['files'] = []
    try:
        tree = ElementTree.fromstring(srr_metadata_url.content)
        for experiment_package in tree.findall('EXPERIMENT_PACKAGE'):
            for run in experiment_package.find('RUN_SET'):
                attributes = run.find('SRAFiles')
                for sra_file in attributes:
                    sra_status = sra_file.attrib['sratoolkit']
                    if sra_status == '1' or 1:
                        file_name = sra_file.attrib['filename']
                        if 'SRR' in file_name:
                            accession = file_name
                            file_path = sra_file.attrib['url']
                            paths_sra[accession]['files'].append(file_path)
                        else:
                             continue
    except:
        paths_sra = None
    return paths_sra

def get_sra_path_from_sra(run_accessions):
    run_accessions = [x for x in run_accessions if x]
    if len(run_accessions) > 100:
        run_lists = list(run_accessions_to_parts(run_accessions))
        result_list = pool_retrieve_xml_from_sra(run_lists)
        paths_sra = result_list
    else:
        paths_sra = retrieve_xml_from_sra(run_accessions)
    return paths_sra

def get_sra_path_from_ena(study_accession, run_accessions):
    paths_sra = {}
    try:
        request_url = f'http://www.ebi.ac.uk/ena/portal/api/filereport?accession={study_accession}&result=read_run&fields=run_accession,sra_ftp'
        sra_results = pd.read_csv(request_url, delimiter='\t')
        if sra_results.shape[0] > 0:
            for i, accession in enumerate(list(sra_results['run_accession'])):
                if accession in run_accessions:
                    paths_sra[accession] = {'files': []}
                    file_path = str(list(sra_results['sra_ftp'])[i])
                    file_path = "ftp://" + file_path
                    paths_sra[accession]['files'].append(file_path)
        else:
            paths_sra = {}
    except:
        paths_sra = {}
    return paths_sra

def get_fastq_path_from_ena(study_accession, run_accessions):
    paths_fastq = {}
    try:
        request_url = f'http://www.ebi.ac.uk/ena/portal/api/filereport?accession={study_accession}&result=read_run&fields=run_accession,fastq_ftp'
        fastq_results = pd.read_csv(request_url, delimiter='\t')
        if fastq_results.shape[0] > 0:
            for i in range(0, len(list(fastq_results['run_accession']))):
                accession = list(fastq_results['run_accession'])[i]
                if accession in run_accessions:
                    paths_fastq[accession] = {'files': []}
                    if ';' in str(list(fastq_results['fastq_ftp'])[i]):
                        # The fastq file paths are split by ";" when retrieved using the above ENA url. If there is no ';' separater,
                        # this means that the number of available fastq files is <= 1. 2 is the minimum number required. If only 1
                        # fastq file is available then an empty dictionary is returned.
                        file_list = str(list(fastq_results['fastq_ftp'])[i]).split(';')
                        for file_path in file_list:
                            file_path = "ftp://" + file_path
                            paths_fastq[accession]['files'].append(file_path)
                    else:
                            paths_fastq = {}
                else:
                    continue
    except:
        paths_fastq = paths_fastq
    return paths_fastq



import pandas as pd
import sys

def get_specimen(xlsx_dict, type):

    specimen_df = xlsx_dict['specimen_from_organism'].merge(
        xlsx_dict[type],
        how="outer",
        on="specimen_from_organism.biomaterial_core.biomaterial_id"
    )

    return specimen_df

def merge_cell_lines(xlsx_dict, merged_df, experimental_design):

    if experimental_design == "cell_line_only":

        merged_df = xlsx_dict['cell_line'].merge(
            merged_df,
            how="outer",
            on="cell_line.biomaterial_core.biomaterial_id"
        )

        specimen_df = get_specimen(xlsx_dict, type="cell_line")
        merged_df = merged_df.merge(
            specimen_df,
            how="outer",
            on="cell_line.biomaterial_core.biomaterial_id"
            )

    elif experimental_design == "organoid_only":

        merged_df = xlsx_dict['organoid'].merge(
            merged_df,
            how="outer",
            on="organoid.biomaterial_core.biomaterial_id"
        )

        specimen_df = get_specimen(xlsx_dict, type="organoid")
        merged_df = merged_df.merge(
            specimen_df,
            how="outer",
            on="cell_line.biomaterial_core.biomaterial_id"
            )

    elif experimental_design == "organoid":

        merged_df = xlsx_dict['organoid'].merge(
            merged_df,
            how="outer",
            on="organoid.biomaterial_core.biomaterial_id"
        )

        merged_df = xlsx_dict['cell_line'].merge(
            merged_df,
            how="outer",
            on="cell_line.biomaterial_core.biomaterial_id"
        )

        specimen_df = get_specimen(xlsx_dict, type="cell_line")
        merged_df = merged_df.merge(
            specimen_df,
            how="outer",
            on="cell_line.biomaterial_core.biomaterial_id"
            )

    return merged_df


import pathlib
from setuptools import setup, find_packages

HERE = pathlib.Path(__file__).parent

README = (HERE / "README.md").read_text()
install_requires = [line.rstrip() for line in (HERE / "requirements.txt").read_text().splitlines()]

setup(
    name="hca2scea",
    version="v0.1.1",
    description="A tool to assist in the automatic conversion of hca metadata to scea metadata MAGE-TAB files.",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/ebi-ait/hca-to-scea-tools",
    author="Ami Day, Yusra Haider, Alegria Aclan, Javier Ferrer",
    author_email="ami@ebi.ac.uk, yhaider@ebi.ac.uk, aaclan@ebi.ac.uk, javier.f.g@um.es",
    classifiers=[
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: Apache Software License",
    ],
    packages=find_packages(exclude=["tests", "tests.*"]),
    install_requires=install_requires,
    include_package_data=True,
    entry_points={
        "console_scripts": [
            "hca2scea=hca_to_scea.hca2scea:main",
        ]
    },
)




from pathlib import Path
import numpy as np
import pandas as pd
import sys
import copy

from hca_to_scea.helpers import utils
from hca_to_scea.helpers import get_protocol_map
from hca_to_scea.helpers import cell_lines

def delete_unused_tabs(xlsx_dict):

    xlsx_dict_tmp = copy.deepcopy(xlsx_dict)
    for tab in xlsx_dict.keys():
        if "protocol" in tab:
            if xlsx_dict[tab].empty:
                del xlsx_dict_tmp[tab]
    return xlsx_dict_tmp

def delete_unused_protocols(xlsx_dict_tmp):

    '''Delete unused protocol keys.'''
    xlsx_dict_tmp2 = copy.deepcopy(xlsx_dict_tmp)
    biomaterials_tabs = ["specimen_from_organism","cell_line","organoid","cell_suspension"]
    for tab in biomaterials_tabs:
        if tab in xlsx_dict_tmp.keys():
            protocol_ids = [protocol_id for protocol_id in xlsx_dict_tmp[tab].columns if ".protocol_core.protocol_id" in protocol_id]
            for protocol_id in protocol_ids:
                if xlsx_dict_tmp[tab][protocol_id].isna().all():
                    del xlsx_dict_tmp2[tab][protocol_id]

    return xlsx_dict_tmp2

def remove_unused_protocols(xlsx_dict):

    '''Delete unused tabs.'''
    xlsx_dict_tmp = delete_unused_tabs(xlsx_dict)
    '''Delete unused protocol keys.'''
    xlsx_dict_tmp2 = delete_unused_protocols(xlsx_dict_tmp)

    return xlsx_dict_tmp2

def clean_dictionary(xlsx_dict):

    for filename in xlsx_dict.keys():
        xlsx_df = xlsx_dict[filename]
        xlsx_df_clean = xlsx_df.replace('', np.nan).dropna(how="all")
        xlsx_df_clean = xlsx_df_clean.loc[:, ~xlsx_df_clean.columns.str.contains('^Unnamed')]
        xlsx_dict[filename] = xlsx_df_clean

    '''Remove NAs in protocol dataframes.'''
    for protocol_type in get_protocol_map.map_of_hca_protocol_type_id_keys.keys():
        xlsx_dict[protocol_type] = xlsx_dict[protocol_type].fillna('')

    return xlsx_dict

def multitab_excel_to_dict(work_dir, excel_file):

    '''Each tab in the multitab excel_file is saved as a
    pandas dataframe. The set of dataframes are stored in
    a dictionary.
    '''

    xlsx = pd.ExcelFile(excel_file, engine='openpyxl')

    xlsx_dict = {}

    for sheet in xlsx.sheet_names:
        path = Path(work_dir)
        path.mkdir(parents=True, exist_ok=True)
        filename = f"{utils.convert_to_snakecase(sheet)}"
        xlsx_df = pd.read_excel(
            excel_file,
            sheet_name=sheet,
            header=0,
            skiprows=[0,1,2,4],
            engine='openpyxl')
        xlsx_dict[filename] = xlsx_df

    xlsx_dict_clean = clean_dictionary(xlsx_dict)

    return xlsx_dict_clean

def merge_sample_types(xlsx_dict: {},experimental_design) -> pd.DataFrame():

    '''process.insdc_experiment.insdc_experiment_accession' is present in multiple tabs. Save the cell_suspension experiment accessions with a unique name.'''
    cell_suspension_column = xlsx_dict['cell_suspension']
    if 'process.insdc_experiment.insdc_experiment_accession' in cell_suspension_column.columns:
        cell_suspension_column["cell_suspension.insdc_experiment.insdc_experiment_accession"] = list(
            cell_suspension_column['process.insdc_experiment.insdc_experiment_accession'])
    else:
        cell_suspension_column["cell_suspension.insdc_experiment.insdc_experiment_accession"] = ['']*len(list(cell_suspension_column["cell_suspension.biomaterial_core.biomaterial_id"]))


    merged_df = cell_suspension_column.merge(
        xlsx_dict['sequence_file'],
        how="outer",
        on="cell_suspension.biomaterial_core.biomaterial_id"
    )

    if experimental_design != "standard":

        """If a cell_suspension id is derived from a cell_line id or an organoid id, get the specimen id."""
        merged_df = cell_lines.merge_cell_lines(xlsx_dict, merged_df, experimental_design)

    else:

        merged_df = xlsx_dict['specimen_from_organism'].merge(
            merged_df,
            how="outer",
            on="specimen_from_organism.biomaterial_core.biomaterial_id"
        )

    """Merge the merged_df with the donor_organism df via the linked
     donor_organism ids."""
    merged_df = xlsx_dict['donor_organism'].merge(
        merged_df,
        how="outer",
        on="donor_organism.biomaterial_core.biomaterial_id"
    )

    return merged_df

def merge_protocol_types(xlsx_dict: {}, merged_df: pd.DataFrame(), experimental_design: str) -> pd.DataFrame():

    """Merge the merged_df with the library_preperation_protocol df via the linked
     library_preparation_protocol ids."""
    merged_df = xlsx_dict['library_preparation_protocol'].merge(
        merged_df,
        how="outer",
        on="library_preparation_protocol.protocol_core.protocol_id"
    )

    """Merge the merged_df with the sequencing_protocol df via the linked
     sequencing_protocol ids."""
    merged_df = xlsx_dict['sequencing_protocol'].merge(
        merged_df,
        how="outer",
        on="sequencing_protocol.protocol_core.protocol_id"
    )

    return merged_df

def merge_dataframes(xlsx_dict: {}, experimental_design: str) -> pd.DataFrame():

    merged_df = merge_sample_types(xlsx_dict, experimental_design)
    merged_df = merge_protocol_types(xlsx_dict, merged_df, experimental_design)

    return merged_df

def merge_rows_by_read_pair(merged_df):

    merged_df_read1 = merged_df.loc[merged_df['sequence_file.read_index'] == 'read1']
    merged_df_read2 = merged_df.loc[merged_df['sequence_file.read_index'] == 'read2']

    merged_df_read2_short = merged_df_read2[[
        'cell_suspension.biomaterial_core.biomaterial_id',
        'sequence_file.file_core.file_name',
        'sequence_file.read_length',
        'sequence_file.lane_index',
    ]]

    merged_df_joined = merged_df_read1.merge(
        merged_df_read2_short,
        on=['cell_suspension.biomaterial_core.biomaterial_id', 'sequence_file.lane_index'],
        suffixes=("_read1", "_read2")
    )

    return merged_df_joined


def merge_index_reads(merged_df, merged_df_by_reads):

    merged_df = merged_df.apply(lambda x: x.str.strip() if x.dtype == "object" else x)

    merged_df_index1 = merged_df.loc[merged_df['sequence_file.read_index'] == 'index1']

    merged_df_index1_short = merged_df_index1[[
        'cell_suspension.biomaterial_core.biomaterial_id',
        'sequence_file.file_core.file_name',
        'sequence_file.read_length',
        'sequence_file.lane_index',
    ]]

    merged_df_index1_short.columns = [f"{x}_index1" for x in merged_df_index1_short.columns]

    merged_df_by_index = merged_df_by_reads.merge(
            merged_df_index1_short,
            left_on=['cell_suspension.biomaterial_core.biomaterial_id', 'sequence_file.lane_index'],
            right_on=["cell_suspension.biomaterial_core.biomaterial_id_index1", 'sequence_file.lane_index_index1'],
        )

    return merged_df_by_index


def reorder_columns(df):

    df.reset_index(inplace=True)
    df = df.rename(
        columns={'sequence_file.file_core.file_name': 'sequence_file.file_core.file_name_read1'})
    df_sorted = df.reindex(sorted(df.columns), axis=1)
    df_sorted = df_sorted.apply(lambda x: x.str.strip() if x.dtype == "object" else x)
    reordered_df = df_sorted
    reordered_df['donor_organism.organism_age'] = df['donor_organism.organism_age']

    return reordered_df


def clean_df(df):

    '''Remove duplicate columns.'''
    df = df[[x for x in df.columns if x not in df.columns[df.columns.duplicated()]]]

    '''Reorder columns by column names.'''
    df_clean = reorder_columns(df)

    return df_clean


def create_new_protocol_columns(df, xlsx_dict, experimental_design):

    '''This extracts the lists from protocol types which can have more than one instance and creates extra columns in the
    df for each of the items.'''
    for (protocol_type, protocol_field) in get_protocol_map.multiprotocols.items():
        if xlsx_dict.get(protocol_type) is not None:
            xlsx_dict[protocol_type] = xlsx_dict[protocol_type].fillna('')
            proto_df, proto_df_columns = get_protocol_map.split_multiprotocols(df, protocol_field)
            for proto_column in proto_df_columns:
                if get_protocol_map.map_of_hca_protocol_type_id_keys.get(protocol_type) == None:
                    get_protocol_map.map_of_hca_protocol_type_id_keys[protocol_type] = []
                get_protocol_map.map_of_hca_protocol_type_id_keys[protocol_type].append(proto_column)

            df = df.merge(proto_df, left_index=True, right_index=True)

    return df